function [alpha,beta,W] = rap_learn(n,modeltype,M,alpha,beta,W,varargin)

%  [alpha,beta,W] = rap_learn(n,modeltype,M,alpha,beta,W,varargin)
%
% learn a harmonium model where
%   n       term/document matrix
%   modeltype 	'2layer_mf' : meanfield version
%               '2layer_sampling' : Gibbs sampling
%   M       number of hidden units
%   alpha   'alpha' parameter or [] if to be initialized
%   beta,W  same as 'alpha'
%
%   varargin can consist of string followed by value
%   
%  LEARNING RATES
%
%  'etaW'               learning rate for 'alpha'
%  'etaBeta','etaW'     same for beta and W
%  'momAlpha'           momentum weight for 'alpha'
%  'momW','momBeta'     same as 'momAlpha'
%  'decayAlpha'         decay Term for 'alpha'
%  'decayBeta','decayW' same as 'decayAlpha'
%
%  LEARNING PARAMTERS
%
%  'bs','batchsize'     batchsize to process each step (default:100)
%  'nU','nUpdates'      # of updates to perform in total (-1 \equiv forever(default))
%  'rB','randomBatches' 1 if batches should be randomized after
%  			every sweep over the training data (default:on)
%  'Mm'                 range of hidden units for binomial model (default:1)
%  'Mn'                 as 'Mm' if observed model is binomial (default:1)
%  'h','hidden'         model for the hidden units 'poisson' or
%			'binom' (default)
%  'o','observed'       model for observed units (default: 'poisson')
%  'l','langevin'       use langevin sampling '1' or not '0' (default:0)
%  'noise'              variance of noise added to 0 counts (default is 0)
%
%  'avgsample' 		# of update steps which should lie between
%  			the parameter updates
%  'navgsamples'	# of samples to average the parameters over
%  'q'
%
%  PLOTS AND SAVEING
%
%  'ws','windowsize'    for plots only, safe to ignore
%  'sfile'              output filename default is 'harmonium_model_' data
%  'updplt'             interval in which figures are updated
%  'lda'                use lda to plot, specify filename with variables
%                       'labels' in them (needs stprtool toolbox)
%  'silent'             1: no plots, store the variables in a file
%                          to be able to have a look at them
%                          offline
%                       0: plot something
% 
% Example: see the script "sampleRun.m" for an example of how this
% function can be called.
%
% Peter Gehler 2005 pgehler@tuebingen.mpg.de

%
% some default values
%

Mh=1;Mm = 1;Mn = 1;batchsize = 100;nUpdates = -1;
etaAlpha = 0.05; mrateAlpha = .95; decayAlpha = 0;
etaBeta = 0.05; mrateBeta = .95; decayBeta = 0;
etaW = 0.05; mrateW = .95; decayW = 0;
langevin = 0;randomBatches = 1;nCdsteps = 1;
windowsize = 100;updplt = 100;navgsamples = 10;
sfile = ['harmonium_model_' date];
hidden = 'binom';observed = 'poisson';
uselda = 0;noise = 0;avgsample = 200;
MAXPLOT = 20000;silent =0;

%
% PARSE INPUT ARGUMENTS
%
if nargin > 6
    for cnt = 1:2:nargin-6
        switch(varargin{cnt})
            case {'bs','batchsize'},batchsize = varargin{cnt+1};
            case {'nU','nUpdates'},nUpdates = varargin{cnt+1};
            case 'etaW',etaW = varargin{cnt+1};
            case 'etaBeta',etaBeta = varargin{cnt+1};
            case 'etaAlpha',etaAlpha = varargin{cnt+1};
            case 'momBeta',mrateBeta = varargin{cnt+1};
            case 'momW',mrateW = varargin{cnt+1};
            case 'momAlpha',mrateAlpha = varargin{cnt+1};
            case 'decayAlpha',decayAlpha = varargin{cnt+1};
            case 'decayBeta',decayBeta = varargin{cnt+1};
            case 'decayW',decayW = varargin{cnt+1};
            case 'nCdsteps',nCdsteps = varargin{cnt+1};
            case {'rB','randomBatches'}, randomBatches = varargin{cnt+1};
            case 'sfile', sfile = varargin{cnt+1};
            case 'updplt', updplt = varargin{cnt+1};
            case 'Mm', Mm = varargin{cnt+1};
            case 'Mn', Mn = varargin{cnt+1};
            case 'Mh', Mh = varargin{cnt+1};
            case {'h','hidden'}, hidden = varargin{cnt+1};
            case {'o','observed'}, observed = varargin{cnt+1};
            case {'l','langevin'}, langevin = varargin{cnt+1};
            case {'ws','windowsize'},windowsize = 100;
            case {'noise'}, noise = varargin{cnt+1};
            case 'avgsample',avgsample = varargin{cnt+1};
            case 'navgsamples', navgsamples = varargin{cnt+1};
            case 'lda',uselda=1; labelfilename = varargin{cnt+1};
            case 'Nh', Nh = varargin{cnt+1};
            case 'Ipermute',Ipermute = varargin{cnt+1};
	 case 'q', q = varargin{cnt+1};
	  case 'silent', silent = varargin{cnt+1};
            otherwise
                error(sprintf('argument %s is not recognized ', varargin{cnt}));
        end
    end
end

%
% PRINT OUT WHAT WE ARE DOING
%
if max(n(:)) == 1 && strcmp('observed','poisson')
    warning('rap_learn','!!!!!Using POISSON MODEL but wordcount is 1 !!!!');
end

if max(n(:)) > 1 && strcmp('observed','binom')
    warning('rap_learn','!!!!!Using BERNOULLI MODEL but wordcount is >1 !!!!');
end

fprintf('using a batchsize of %d\n',batchsize);
fprintf('Observed Model : ');
if strcmp(observed,'poisson'), fprintf('Poisson\n');
elseif strcmp(observed,'binom'), fprintf('Binomial with %d counts\n',Mn); end
fprintf('Hidden Model : ');
if strcmp(hidden,'poisson'), fprintf('Poisson\n');
elseif strcmp(hidden,'binom'), fprintf('Binomial with %d counts\n',Mm); end
fprintf('will sample %d variables with a distance %d\n',navgsamples,avgsample);

fprintf('using %d steps of Contrastive DIvergence\n',nCdsteps);
fprintf('saving data in ''%s''\n',sfile);


%
% PLOT PROJECTION WITH  LINEAR DISCRIMINANT ANALYSIS 
%
if uselda,
    ll = load(labelfilename);
    if length(ll)~=1, error(['store only labels in ' labelfilename]);end
    ll_name = fieldnames(ll);
    labels = getfield(ll,ll_name{1});
    if size(labels,2) ~= size(n,2)
	labels = labels';
	if size(labels,2) ~= size(n,2)
	    size(labels)
	    size(n)
	    error('wrong dimensions of labels');
	end
    end
end


[D,N] = size(n); % D = #documents, N = #words

% initialize alpha to meet the average word count
if ~prod(size(alpha))
    if noise
        b = n == 0;        
        nn = n + noise*rand(size(n)).*b;
        avgn = sum(nn,2)/N;
        clear b nn
    else
        avgn = sum(n,2)/N;
    end
    
    if strcmp(observed,'poisson')
        alpha = log(avgn);
    else
        alpha = -log(Mn./avgn-1);
    end
end

%
% W is connection matrix between observed / 1st hidden layer
%
if ~prod(size(W)),
    W = randn(M,D)./(M*D*max(abs(n(:))));
end %4/max(sum(n));end
if ~prod(size(beta)), 
    beta = .01 * randn(M,1);
end

if (size(W,2) ~= size(alpha,1)) || size(W,1) ~= size(beta,1)
    error('dimension mismatch');
end

%
% PREPARE VARIABLES AND SOME LEARNING PARAMETERS
%
if nUpdates < 0, nUpdates = 1e100; end % update "forever"

initialize_auxiliaries;

%
% randomize the training data (indices)
%
if randomBatches, ind = randperm(N); 
else ind = 1:N; end
cnt = 0;

%
% add a bit of noise to the wordcounts (not used any more)
%
if noise,
    nzeros = n == 0;
    n_noise = n + noise*rand(size(nzeros)).*nzeros;
else,n_noise = n;end

%
% instead of one sample as output average the parameters over
% 'navgsample' samples with a distance of 'avgsamples' between them.
%
if avgsample
    samplecount = 0;
    W_avg{1} = 0; W_avg = repmat(W_avg,navgsamples,1);
    beta_avg = W_avg; alpha_avg = W_avg;
end

%
% save every minute, store current time
%
t_lastsaved = clock;

%
% PARAMETER UPDATE LOOP
%
for upd = 1:nUpdates

  % 
  % get next training batch or rerandomize
  %
    if (cnt+1)*batchsize>N
        t_ind = ind(cnt*batchsize+1:end);
        t_ind = [t_ind,ind(1:batchsize-size(t_ind,2)-1)];
        cnt = 0;
        if randomBatches, ind = randperm(N); end
        if noise
            n_noise = n + noise*rand(size(nzeros)).*nzeros;
        else
            n_noise = n;
        end
    else
        t_ind = ind(cnt*batchsize+1:(cnt+1)*batchsize);
        cnt = cnt + 1;
    end
    n0 = n_noise(:,t_ind);

    %
    % these variables are used more often so produce them once only
    %
    Alpha = repmat(alpha,1,size(n0,2));
    Beta = repmat(beta,1,size(n0,2));

    %
    % GENERATE SAMPLES FOR PARAMETER UPDATES
    %
    switch lower(modeltype)
        case '2layer_mf' %%%%%%%%% 2 LAYER MEANFIELD %%%%%%%%%%%%
            [n1,m0] = meanfield(n0,Alpha,Beta,W,observed,hidden,Mm,Mn,1);
            m1 = Mm * sigmoid(Beta+W*n1);
        case '2layer_sampling' %%%%%%%%% 2 LAYER SAMPLING %%%%%%%%%%%%
            [n1,p] = gibbs_sampler(n0,alpha,beta,W,observed,hidden,nCdsteps,Mm,Mn);
            m0 = Mm * p;
            m1 = Mm * sigmoid(Beta+W*n1);
        otherwise
            error('no such modeltype');
    end

    %
    % UPDATE PARAMETERS
    %
    if etaAlpha
        dAlpha = etaAlpha *(mean(n0,2)-mean(n1,2) -decayAlpha*alpha) +mrateAlpha*dAlpha;
        if sum(isnan(dAlpha(:))), dAlpha(isnan(dAlpha)) = 0;end
        alpha = alpha + dAlpha; 
	% langevin update : 
       %alpha = alpha + dAlpha + sqrt(2*etaAlpha/batchsize)*randn(size(alpha));
    end
    if etaBeta
        dBeta = etaBeta*(mean(m0,2)-mean(m1,2)-decayBeta*beta)+mrateBeta*dBeta;
        if sum(isnan(dBeta(:))), dBeta(isnan(dBeta)) = 0;end
        beta = beta + dBeta;
	% langevin update : 
        %beta = beta + dBeta + sqrt(2*etaAlpha/batchsize)*randn(size(beta));
    end
    if etaW
        dW = etaW*( (m0*n0')/batchsize - (m1*n1')/batchsize -decayW*W)+mrateW*dW;
        if sum(isnan(dW(:))), dW(isnan(dW)) = 0;end
        W = W + dW;
	% langevin update : 
        %W = W + dW + sqrt(2*etaW/batchsize)*randn(size(W));
    end

    
    %
    % UPDATE VARIABLES FOR PLOTTING 
    %
    update_auxiliaries;

    %
    % REDRAW THE PLOTS
    %
    if ~silent
      update_plots;
    end
    
    
    
    %
    % store the parameters in the list to be able to compute the
    % average later on
    %
    if avgsample && ~mod(upd,avgsample)        
        W_avg = shift(W_avg,1); W_avg{1} = W;
        alpha_avg = shift(alpha_avg,1); alpha_avg{1} = alpha;
        beta_avg = shift(beta_avg,1); beta_avg{1} = beta;
        samplecount = samplecount + 1;
    end
    
    
    %
    % SAVE THE PARAMETER every half hour
    %
    if etime(clock,t_lastsaved) > 3000
      % compute the average
        if avgsample & length(W_avg)
            al = alpha; bt= beta; w = W;
            [alpha,beta,W] = avgSample(samplecount,W_avg,alpha_avg,beta_avg);
        end
        save(sfile,'alpha','beta','W',saveparam);
	save(sfile,'etaAlpha','etaBeta','etaW','-append',saveparam);
        save(sfile,'modeltype','-append',saveparam);
        if strcmp(observed,'binom'), save(sfile,'Mn','-append',saveparam);end
        if strcmp(hidden,'binom'), save(sfile,'Mm','-append',saveparam);end
        if exist('Ipermute','var'), save(sfile,'Ipermute','-append',saveparam);end
        if exist('q','var'), save(sfile,'q','-append',saveparam);end
        
        if avgsample
            alpha = al;beta = bt; W = w;
        end
        t_lastsaved = clock;
        fprintf('%d iterations: saved the variables\n',upd);
	
	if silent
	  update_plots;
	end
	
    end
    
  
end

%
% compute the average
%
if avgsample
    [alpha,beta,W] = avgSample(samplecount,W_avg,alpha_avg,beta_avg);
    %W = correctSignW(W); 
end

fprintf('\n');


% [alpha,beta,W] = avgSample(samplecount,W_avg,alpha_avg,beta_avg);
%
% Compute the average of the parameters. Every 'samplecount'
% iterations the parameters are stored and this function does
% compute the average of those. 
function [alpha,beta,W] = avgSample(samplecount,W_avg,alpha_avg,beta_avg);

samplecount = min(samplecount,length(W_avg));

W = zeros(size(W_avg{1}));
alpha = zeros(size(alpha_avg{1}));
beta = zeros(size(beta_avg{1}));

for i=1:samplecount
    W = W + W_avg{i};
    alpha= alpha + alpha_avg{i};
    beta = beta + beta_avg{i};
end


W = W/samplecount;
alpha = alpha/samplecount;
beta = beta/samplecount;

return;

% correct W such that the vectors all point into the same halfspace.
% Therefore the latent space is ensured to be covered only once and not
% twice
function W = correctSignW(W)

t = sum(W,2)>0;
W(t,:) = - W(t,:);
return
function a = vectorize(a)
a = a(:);
return